#include "skyLaunch.h"

//Lift Coefficient function based on angle of attack
void CoeffFunc (double aoa, double M, double Re, double *cl, double *cm, double *cd)
{
	const int nlift = 11;
	const double factor=0.0;
	static const double AOA[nlift] =
		{-180*RAD,-160*RAD,-150*RAD,-120*RAD,-90*RAD,0*RAD,90*RAD,120*RAD,150*RAD,160*RAD,180*RAD};
	static const double CL[nlift]  = {0.0,-0.3,-0.425,-0.215,0.0,0.0,0.0,0.215,0.425,0.3,0.0};
	static const double CM[nlift]  = {0.0,0.004,0.006,0.012,0.015,0.0,-0.015,-0.012,-0.006,-0.004,0.};
	static const double CD[nlift]  = {1.6,1.4,1.0,0.6,0.75,0,0.75,0.6,1.0,1.4,1.6};
	static double SCL[nlift-1];
	static double SCM[nlift-1];
	static double SCD[nlift-1];
	for(int j = 0; j < nlift-1; j++){
		SCL[j]= (CL[j+1]-CL[j])/(AOA[j+1]-AOA[j]);
		SCM[j]= (CM[j+1]-CM[j])/(AOA[j+1]-AOA[j]);
		SCD[j]= (CD[j+1]-CD[j])/(AOA[j+1]-AOA[j]);
	}
	for (int i = 0; i < nlift-1 && AOA[i+1] < aoa; i++);
	*cl = (CL[i] + (aoa-AOA[i])*SCL[i]);
	*cm = factor*(CM[i] + (aoa-AOA[i])*SCM[i]);
	*cd = (CD[i] + (aoa-AOA[i])*SCD[i]);
}

//Constructor
saturnv21::saturnv21(OBJHANDLE obj, int fmodel):VESSEL2( obj,fmodel)
{
	//meshes
	stage1Mesh = oapiLoadMeshGlobal("int-21/sat5stg1");
	interstage = oapiLoadMeshGlobal("int-21/sat5intstg");
	stage2Mesh = oapiLoadMeshGlobal("int-21/sat5stg2");
	fairingMesh[0] = NULL;
	fairingMesh[1] = NULL;

	//Options from files
	launchHDG = 73;
	desiredApogee = 300;
	desiredPerigee = 280;
	status = 0;
	payloadWeight = 77500;
	//Visual handlers
	stage1 = intstg = stage2 = fairing1 = fairing2 = MESH_UNDEFINED;

	//fuel tanks
	ph_1 = NULL;
	ph_2 = NULL;

	//Timings
	goForSkirtSep = false;
	goForFairing = false;
	waitSafeDistance = false;
	goForDeorbit = false;

}

void saturnv21::clbkSetClassCaps(FILEHANDLE cfg){

}

//Loafing options from scenario file
void saturnv21::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
	fairing = true;
	char *line = NULL;

	while(oapiReadScenario_nextline(scn,line)){

		if(!strnicmp(line, "CONFIGURATION", 13))
			sscanf(line+13, "%d", &status);

		else if(!strnicmp(line, "PAYLOAD_WEIGHT", 14))
			sscanf(line+14, "%lf", &payloadWeight);

		else if(!strnicmp(line, "NOFAIR", 6))
			fairing = false;

		else
			ParseScenarioLineEx(line, vs);
	}
	
	if (fairing){
		fairingMesh[0] = oapiLoadMeshGlobal("int-21/sat5sky1");
		fairingMesh[1] = oapiLoadMeshGlobal("int-21/sat5sky2");
	}

	switch(status){
	case 0:
		setLaunchConfiguration(); //before liftoff
		break;
	case 1:
		setFlyingConfiguration(); //after liftoff running on 1st stage
		break;
	case 2:
		setStage2Configuration(); //after liftoff running on 2nd stage
		break;
	}

}

//Loading option for a dynamically created vessel
void saturnv21::clbkSetStateEx (const void *st)
{
	DefSetStateEx (st);
	setLaunchConfiguration();
}


//Saves the the state into scenario file
void saturnv21::clbkSaveState (FILEHANDLE scn)
{
	VESSEL2::clbkSaveState(scn);
	
	oapiWriteScenario_int(scn, "CONFIGURATION", status);

	oapiWriteScenario_float(scn, "PAYLOAD_WEIGHT", payloadWeight);

	if(!fairing)
		oapiWriteLine(scn, "NOFAIR");

}

//Launch configuration with all stages
void saturnv21::setLaunchConfiguration()
{
	//****************Physical parameters************
	SetSize(10);
	SetTouchdownPoints(_V(0,-4.5,-60.2), _V(-4.5,0,-60.2), _V(4.5,0,-60.2));
	SetSurfaceFrictionCoeff(0.5,0.5);
	SetEmptyMass(135200 + 1905 + 39050 + payloadWeight); //1ststage, interstage, 2nd stages,  and payload masses
	CreateAirfoil (LIFT_VERTICAL,   _V(0,0,-0.5), CoeffFunc, 50, 0, 0);
	SetRotDrag(_V(0.005,0.005,0.005));
	//***********************************************

	//***************Fuel tanks**********************
	if(!ph_1) ph_1 = CreatePropellantResource(2151000);
	if(!ph_2) ph_2 = CreatePropellantResource(451830);
	SetDefaultPropellantResource (ph_1);

	//Attachment definition
	if (!payload_attach) payload_attach = CreateAttachment (false, _V(0,0,63), _V(0,0,1), _V(0,1,0), "X");
	
	//1st stage engines and main group
	//***********************************************
	th_f1[0] = CreateThruster(_V(0,0,-21.5), _V(0,0,1), 7740000, ph_1, 2982.24, 2599.65);
	th_f1[1] = CreateThruster(_V(3.2,3.2,-21.5), _V(0,0,1), 7740000, ph_1, 2982.24, 2599.65);
	th_f1[2] = CreateThruster(_V(3.2,-3.2,-21.5), _V(0,0,1), 7740000, ph_1, 2982.24, 2599.65);
	th_f1[3] = CreateThruster(_V(-3.2,-3.2,-21.5), _V(0,0,1), 7740000, ph_1, 2982.24, 2599.65);
	th_f1[4] = CreateThruster(_V(-3.2,3.2,-21.5), _V(0,0,1), 7740000, ph_1, 2982.24, 2599.65);

	thg_main = CreateThrusterGroup(th_f1, 5, THGROUP_MAIN);
	
	//Exhaust and Contrail
	SURFHANDLE tex_f1 = oapiRegisterExhaustTexture ("Exhaust");
	//SURFHANDLE cont_f1 = oapiRegisterExhaustTexture ("Contrail2");
	int i;
	for(i = 0 ; i < 5; i++) AddExhaust(th_f1[i], 30, 2, tex_f1);
	//for(i = 0; i < 5; i++) AddExhaust(th_f1[i], 50, 50, cont_f1);

	//visual
	SetCameraOffset(_V(0,0,85));
	stage1 = AddMesh(stage1Mesh, &OFS_S1);
	intstg = AddMesh(interstage, &OFS_INTERSTAGE);
	stage2 = AddMesh(stage2Mesh, &OFS_S2);

	SetMeshVisibilityMode(stage1, MESHVIS_ALWAYS);
	SetMeshVisibilityMode(intstg, MESHVIS_ALWAYS);
	SetMeshVisibilityMode(stage2, MESHVIS_ALWAYS);

	ShiftCG(_V(0,0,10));
}

//Saturn is flying on 1st stage
void saturnv21::setFlyingConfiguration(){
	setLaunchConfiguration();
	SetTouchdownPoints(_V(0,-4.5,-20), _V(-4.5,0,-20), _V(4.5,0,-20));
}


//Saturn is flying on 2nd stage
void saturnv21::setStage2Configuration(){

	//****************Physical parameters************
	SetSize(10);
	SetSurfaceFrictionCoeff(0.5,0.5);
	SetEmptyMass(1905 + 39050 + payloadWeight); //interstage, 2nd stages,  and payload masses
	CreateAirfoil (LIFT_VERTICAL,   _V(0,0,-0.5), CoeffFunc, 50, 0, 0);
	SetRotDrag(_V(0.005,0.005,0.005));
	//***********************************************

	//***************Fuel tanks**********************
	if(!ph_2) ph_2 = CreatePropellantResource(451830);
	SetDefaultPropellantResource (ph_2);

	//Attachment definition
	if (!payload_attach) payload_attach = CreateAttachment (false, _V(0,0,39), _V(0,0,1), _V(0,1,0), "X");
	
	//2nd stage engines and main group
	//***********************************************
	th_j2[0] = CreateThruster(_V(0,0,-3.5), _V(0,0,1), 1033000, ph_2, 4130, 1962);
	th_j2[1] = CreateThruster(_V(1.5,1.5,-3.5), _V(0,0,1), 1033000, ph_2, 4130, 1962);
	th_j2[2] = CreateThruster(_V(1.5,-1.5,-3.5), _V(0,0,1), 1033000, ph_2, 4130, 1962);
	th_j2[3] = CreateThruster(_V(-1.5,-1.5,-3.5), _V(0,0,1), 1033000, ph_2, 4130, 1962);
	th_j2[4] = CreateThruster(_V(-1.5,1.5,-3.5), _V(0,0,1), 1033000, ph_2, 4130, 1962);

	thg_main = CreateThrusterGroup(th_j2, 5, THGROUP_MAIN);
	CreateStage2RCS();

	//Exhaust and Contrail
	SURFHANDLE tex_j2 = oapiRegisterExhaustTexture ("Exhaust");
	//SURFHANDLE cont_j2 = oapiRegisterExhaustTexture ("Contrail2");
	int i;
	for(i = 0 ; i < 5; i++) AddExhaust(th_j2[i], 10, 0.8, tex_j2);
	//for(i = 0; i < 5; i++) AddExhaust(th_j2[i], 50, 50, cont_f1);
	

	//Timings
	sepCountdown = oapiGetSimTime();
	goForSkirtSep = true;
	goForFairing = true;

	//Visual
	SetCameraOffset(_V(0,0,60));
	const VECTOR3 ofs = {0,0,13.3};
	if (intstg == MESH_UNDEFINED) intstg = AddMesh(interstage);
	if (stage2 == MESH_UNDEFINED) stage2 = AddMesh(stage2Mesh, &ofs);

	ShiftCG(_V(0,0,15));
}


void saturnv21::clbkPostStep (double simt, double simdt, double mjd){
	
	if(!status && GetEngineLevel(ENGINE_MAIN) > 0){
		status = 1;
		SetTouchdownPoints(_V(0,-4.5,-20), _V(-4.5,0,-20), _V(4.5,0,-20));
	}

	if(status < 2 && oapiGetPropellantMass(ph_1) < 1)
		discardStage1();

	if(status < 2 && oapiGetPropellantMass(ph_1) < 107550){
		SetThrusterLevel(th_f1[0], 0.2);
		SetThrusterLevel(th_f1[1], 0.2);
		SetThrusterLevel(th_f1[2], 0.2);
		SetThrusterLevel(th_f1[3], 0.2);
		SetThrusterLevel(th_f1[4], 0.2);
	}

	if(goForSkirtSep && (oapiGetSimTime() - sepCountdown > skirtSepInterval)){
		skirtSep();
		goForSkirtSep = false;
	}

	if(goForFairing && (oapiGetSimTime() - sepCountdown > fairingInterval)){
		fair();
		goForFairing = false;
	}

	if(waitSafeDistance && (oapiGetSimTime() - sepCountdown > 10)){
		ActivateNavmode(NAVMODE_RETROGRADE);
		waitSafeDistance = false;
		goForDeorbit = true;
	}

	if (goForDeorbit && (oapiGetSimTime() - sepCountdown > stage2DeorbitInterval)){
		double perigee;
		SetThrusterLevel(th_j2[0], 0.3);
		SetThrusterLevel(th_j2[1], 0.3);
		SetThrusterLevel(th_j2[2], 0.3);
		SetThrusterLevel(th_j2[3], 0.3);
		SetThrusterLevel(th_j2[4], 0.3);

		GetPeDist(perigee);
		if(perigee < 6381000 || oapiGetPropellantMass(ph_2) < 1){
			SetThrusterLevel(th_j2[0], 0);
			SetThrusterLevel(th_j2[1], 0);
			SetThrusterLevel(th_j2[2], 0);
			SetThrusterLevel(th_j2[3], 0);
			SetThrusterLevel(th_j2[4], 0);

			DeactivateNavmode(NAVMODE_RETROGRADE);
			goForDeorbit = false;
		}
	}
}

int  saturnv21::clbkConsumeBufferedKey (DWORD key, bool down, char *kstate){

	if(!down){
		if (status < 2){
			SetThrusterDir(th_f1[1], _V(0,0,1));
			SetThrusterDir(th_f1[2], _V(0,0,1));
			SetThrusterDir(th_f1[3], _V(0,0,1));
			SetThrusterDir(th_f1[4], _V(0,0,1));
		}
		return 0;
	}

	switch(key){
		
	case OAPI_KEY_NUMPAD8:
		if (status < 2){
			SetThrusterDir(th_f1[1], _V(0,0.005,1));
			SetThrusterDir(th_f1[2], _V(0,0.005,1));
			SetThrusterDir(th_f1[3], _V(0,0.005,1));
			SetThrusterDir(th_f1[4], _V(0,0.005,1));
			return 1;
		}
		return 0;
		
	case OAPI_KEY_NUMPAD2:
		if (status < 2){
			SetThrusterDir(th_f1[1], _V(0,-0.005,1));
			SetThrusterDir(th_f1[2], _V(0,-0.005,1));
			SetThrusterDir(th_f1[3], _V(0,-0.005,1));
			SetThrusterDir(th_f1[4], _V(0,-0.005,1));
			return 1;
		}
		return 0;

	case OAPI_KEY_NUMPAD1:
		if (status < 2){
			SetThrusterDir(th_f1[1], _V(0.002,0,1));
			SetThrusterDir(th_f1[2], _V(0.002,0,1));
			SetThrusterDir(th_f1[3], _V(0.002,0,1));
			SetThrusterDir(th_f1[4], _V(0.002,0,1));
			return 1;
		}
		return 0;

	case OAPI_KEY_NUMPAD3:
		if (status < 2){
			SetThrusterDir(th_f1[1], _V(-0.002,0,1));
			SetThrusterDir(th_f1[2], _V(-0.002,0,1));
			SetThrusterDir(th_f1[3], _V(-0.002,0,1));
			SetThrusterDir(th_f1[4], _V(-0.002,0,1));
			return 1;
		}
		return 0;

	case OAPI_KEY_NUMPAD4:
		if (status < 2){
			SetThrusterDir(th_f1[1], _V(-0.02,0,1));
			SetThrusterDir(th_f1[2], _V(0,0.02,1));
			SetThrusterDir(th_f1[3], _V(0.02,0,1));
			SetThrusterDir(th_f1[4], _V(0,-0.02,1));
			return 1;
		}
		return 0;

	case OAPI_KEY_NUMPAD6:
		if (status < 2){
			SetThrusterDir(th_f1[1], _V(0,-0.02,1));
			SetThrusterDir(th_f1[2], _V(-0.02,0,1));
			SetThrusterDir(th_f1[3], _V(0,0.02,1));
			SetThrusterDir(th_f1[4], _V(0.02,0,1));
			return 1;
		}
		return 0;

	case OAPI_KEY_J:
		if(status > 1){
			OBJHANDLE attachee = GetAttachmentStatus(payload_attach);
			DetachChild(payload_attach, 1);
			
			if (attachee){
				oapiSetFocusObject(attachee);
				SetEnableFocus(false);
			}

			sepCountdown = oapiGetSimTime();
			waitSafeDistance = true;
		}
		return 1;
	}

	return 0;
}

//This method separates the 1st stage from the rest
void saturnv21::discardStage1(void){
	status = 2;

	//Creation of the Stage1 Vessel
	VESSELSTATUS2 vs;
	VESSELSTATUS2::FUELSPEC fuel;
	VESSELSTATUS2::THRUSTSPEC thrust;
	memset (&vs, 0, sizeof(vs));
	vs.version = 2;
	GetStatusEx (&vs);
	vs.flag = VS_FUELLIST | VS_THRUSTLIST;
	vs.fuel = &fuel;
	vs.nfuel = 1;
	vs.fuel->idx = 0;
	vs.thruster = &thrust;
	vs.nthruster = 1;
	vs.thruster->idx = 0;
	vs.status = 0;
	Local2Rel(_V(0,0,-10), vs.rpos);
	oapiCreateVesselEx ("Stage-1", "int-21/sat5stg1", &vs);
	
	//Deletion of stage1 from the main vessel
	ClearThrusterDefinitions();
	SetDefaultPropellantResource (ph_2);
	DelPropellantResource(ph_1);

	//Deletion of 1st stage visual part
	DelMesh(stage1);
	ShiftCG(_V(0,0,13));
	setStage2Configuration();

	ActivateNavmode(NAVMODE_KILLROT);
	SetThrusterLevel(th_j2[0], 1);
	SetThrusterLevel(th_j2[1], 1);
	SetThrusterLevel(th_j2[2], 1);
	SetThrusterLevel(th_j2[3], 1);
	SetThrusterLevel(th_j2[4], 1);
}

void saturnv21::skirtSep(){
	//Creation of the skirt Vessel
	VESSELSTATUS2 vs;
	VESSELSTATUS2::FUELSPEC fuel;
	VESSELSTATUS2::THRUSTSPEC thrust;
	memset (&vs, 0, sizeof(vs));
	vs.version = 2;
	//******
	GetStatusEx (&vs);
	//******
	vs.flag = VS_FUELLIST | VS_THRUSTLIST;
	vs.fuel = &fuel;
	vs.nfuel = 1;
	vs.fuel->idx = 0;
	vs.thruster = &thrust;
	vs.nthruster = 1;
	vs.thruster->idx = 0;
	vs.status = 0;
	Local2Rel(_V(0,0,-14.5), vs.rpos);
	oapiCreateVesselEx ("Interstage", "int-21/sat5intstg", &vs);

	SetEmptyMass(39050 + payloadWeight);
	DelMesh(intstg);
}

//This function will be completed in the next release
void saturnv21::fair(){
	//Creation of the fairings Vessel
	VESSELSTATUS2 vs;
	VESSELSTATUS2::FUELSPEC fuel;
	VESSELSTATUS2::THRUSTSPEC thrust;
	memset (&vs, 0, sizeof(vs));
	vs.version = 2;
	GetStatusEx (&vs);
	vs.flag = VS_FUELLIST | VS_THRUSTLIST;
	vs.fuel = &fuel;
	vs.nfuel = 1;
	vs.fuel->idx = 0;
	vs.thruster = &thrust;
	vs.nthruster = 1;
	vs.thruster->idx = 0;
	vs.status = 0;
	//oapiCreateVesselEx ("Stage-1", "Stage1", &vs);
}

void saturnv21::CreateStage2RCS()
{
	// no linear attitude modes are defined at this stage
	THRUSTER_HANDLE th_rot[2];
	th_rot[0] = CreateThruster (_V(0,0, 12), _V(0, 1,0), 10000, ph_2, 5000, 4000);
	th_rot[1] = CreateThruster (_V(0,0,-12), _V(0,-1,0), 10000, ph_2, 5000, 4000);
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_PITCHUP);
	th_rot[0] = CreateThruster (_V(0,0, 12), _V(0,-1,0), 10000, ph_2, 5000, 4000);
	th_rot[1] = CreateThruster (_V(0,0,-12), _V(0, 1,0), 10000, ph_2, 5000, 4000);
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_PITCHDOWN);
	th_rot[0] = CreateThruster (_V( 5,0,0), _V(0, 1,0), 10000, ph_2, 5000, 4000);
	th_rot[1] = CreateThruster (_V(-5,0,0), _V(0,-1,0), 10000, ph_2, 5000, 4000);
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_BANKLEFT);
	th_rot[0] = CreateThruster (_V( 5,0,0), _V(0,-1,0), 10000, ph_2, 5000, 4000);
	th_rot[1] = CreateThruster (_V(-5,0,0), _V(0, 1,0), 10000, ph_2, 5000, 4000);
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_BANKRIGHT);
	th_rot[0] = CreateThruster (_V(0,0, 10), _V( 1,0,0), 10000, ph_2, 5000, 4000);
	th_rot[1] = CreateThruster (_V(0,0,-10), _V(-1,0,0), 10000, ph_2, 5000, 4000);
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_YAWRIGHT);
	th_rot[0] = CreateThruster (_V(0,0, 10), _V(-1,0,0), 10000, ph_2, 5000, 4000);
	th_rot[1] = CreateThruster (_V(0,0,-10), _V( 1,0,0), 10000, ph_2, 5000, 4000);
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_YAWLEFT);
}
